 
#define	DRIVER_NAME	"3DFX Banshee chipset driver"
#define	DRIVER_REV	"$Revision: 1.3 $"

#include <kgi/system.h>
#include <kgi/module.h>
#include "IBM/vga_io.inc"
#include "3dfx/banshee.h"


static void chipset_dump_state(struct  statetable *state) ; 
static void chipset_compute_mode(struct kgi_display *dpy, 
				 struct kgi_mode *mode,
				 struct statetable *setup) ; 


static struct statetable default_state =
{
        0x00,                                           /* MISC */
        {
                0x03, 0x00, 0x00, 0x00, 0x00                    /* SEQ  */
        },
        {
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* CRT  */
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 0x1x */
                0x00
        },
        {
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* GRC  */
                0xFF
        },
        {
                0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, /* ATC  */
                0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
                0x00, 0x00, 0x0F, 0x00, 0x00
        },
        {
                0x00, 0x00 /* ExtVga */
        }
};

#define TEXT_LEN 0x8000  /* 64k framebuffer for text modes */

static struct kgi_chipset banshee_chipinfo =
{
  "3DFX",		        /* Manufacturer		*/
  "Banshee",		        /* Model	(DETECTED) */
  GM_TEXT16 | GM_8BIT | GM_16BIT | GM_24BIT | GM_32BIT, /* Available modes */
  0,			/* Flags		*/
  { 2048, 2048 },	/* Maximum resolution	*/
  NULL,			/* Private data hook	*/
  0x0,			/* addressable RAM (DETECTED) */ 
  { 0, BANSHEE_MAX_CLOCK }		/* DCLK limits (DETECTED) */
};


/* The Memory Bank 0  
   0x0000000 - 0x007FFFF    I/O register remap
   0x0080000 - 0x00FFFFF    CMD/AGP transfer/Mist registers
   0x0100000 - 0x1FFFFFF    2D registers
            ...
   0x0C00000 - 0x0FFFFFF    YUV planar plane
   */
const static struct mem_region banshee_memory_bank_0 =
{
	MEM_NULL,
	0x0,           /*  framebuffer (autodetect) */
	0x0,           /* mem size (autodetect) */
	-1,            /* Decode all bits */
	"Banshee linear framebuffer"
};

/* The Memory Bank 1  
   0x0000000 - 0x1FFFFFF    Linear Frame Buffer
 */
const static struct mem_region banshee_memory_bank_1 =
{
	MEM_NULL,
	0x0,           /* linear framebuffer (autodetect) */
	0x0,           /* mem size (autodetect) */
	-1,            /* Decode all bits */
	"Banshee linear framebuffer"
};


/* The I/O space
   0x0000000 - 0x1FFFFFF    Linear Frame Buffer
 */
const static struct io_region banshee_io_bank =
{
	IO_NULL,
	0x0,           /* linear framebuffer (autodetect) */
	0x0,           /* mem size (autodetect) */
	0xFFFF,        /* Decode all bits */
	"Banshee linear framebuffer"
};


/*
 * 
 * banshee_framebuffer_mmio is not used (see MMIO): TO BE REMOVED 
 *
 */
const static struct kgi_mmio_region banshee_framebuffer_mmio = {
	NULL,                   	/* Next			*/
	"Banshee MMIO framebuffer",	/* Name         	*/
	MMAP_FRAMEBUFFER,		/* Type (Linear)        */
	MMIO_RDWR | MMIO_SHARED,	/* Protection map       */

	0, 	   	                /* Total size of area (DETECTED) */
	0,           	         	/* Physical size       (DETECETED) */
	0,   	         	        /* bus address of the windows */
	0,                      	/* Logical adress       */

	NULL,                  		/* Set offset function  */
	NULL                    	/* private              */
};   

                                                                            
#ifndef PCI_VENDOR_ID_3DFX_BANSHEE
#define PCI_VENDOR_ID_3DFX_BANSHEE 0x0003
#endif

int adjust_virtual(struct kgi_mode *mode)
{
  /* !!! Place here what's necessary */
  return 0;
}


static void text16_set_font(struct kgi_display *dpy, 
			    struct kgi_font *font,kgi_uint from,kgi_uint to)
{
  int i,j; kgiu16 *fbuf; kgiu8 *data;
  
  if (font == NULL)
		return;

  fbuf = ((kgiu16 *)dpy->mode.fb)+0x10000+(from<<5);
  data = font->data + font->dy * from;

  for (i = from; i <= to; i++) {
    for (j=0;j<font->dy;j++)
      *(fbuf++)=*(data++);
    for (;j<32;j++)
      *(fbuf++)=0;
  }

}


static void roundmode(kgi_modereq *gm,int roundx,int roundy)
{
  gm->visible.x = ((gm->visible.x + roundx - 1) / roundx) * roundx;
  gm->visible.y = ((gm->visible.y + roundy - 1) / roundy) * roundy;
  gm->virt.x    = ((gm->virt.x + roundx - 1) / roundx) * roundx;
  gm->virt.y    = ((gm->virt.y + roundy - 1) / roundy) * roundy;
  
  if (gm->visible.x > gm->virt.x)
    gm->virt.x = gm->visible.x;
  if (gm->visible.y > gm->virt.y)
    gm->virt.y = gm->visible.y;
}


void 
kgim_chipset_done(struct kgi_display *dpy)
{
  GET_PRIVATE ; 

  NOTICE(" ENTER kgim_chipset_done");

  if (IOBASE)
    io_free_region(&IOBANK);
  if (IOMEMBASE)
    mem_free_region(&BANK0) ;
  if (MEMBASE)
    mem_free_region(&BANK1) ;
  
  kgi_free_small(private) ; 
  KGIM_PRIVATE(dpy, chipset) = NULL;

  NOTICE(DRIVER_NAME " rev " DRIVER_REV " removed.");
}
	
static void 
adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16)
		? mode->request.dpp.x : 8;
    
#define ADJUST(X) mode->x.X -= (mode->x.X % foo)
    
	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);
    
	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;
    
	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;
    
#undef ADJUST
} 

static void 
chipset_setup_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	int bpp=0;
	GET_PRIVATE;
    
	/* By default, setup for graphics modes */

	dpy->set_font=0;
	switch (mode->request.graphtype) {
	case KGIGT_TEXT16:
	  mode->fb	= MEMBASE;
	  mode->fb_size	= TEXT_LEN;
	  mode->fsize	= mode->request.virt.x * mode->request.virt.y * 2;
	  mode->mem	= &text16_color_mem;
	  mode->io	= &text16_color_io;
	  mode->mmio    = &MMIO;
	  dpy->set_font=text16_set_font;
	  return; 	/* Done setting up */
	case KGIGT_8BIT:
	  bpp = 8;
	  break;
	case KGIGT_16BIT:
	  bpp = 16;
	  break;
	case KGIGT_24BIT:
	  bpp = 24;
	  break;
	case KGIGT_32BIT:
	  bpp = 32;
	  break;
	default:
	  INTERNAL_ERROR;
	  return;
	}

	mode->fb = MEMBASE ;
	mode->fb_size = MEMSIZE ;
	mode->fsize = (mode->request.virt.x * 
	               mode->request.virt.y * 8)/bpp;
	mode->mem	= NULL ;
	mode->io	= NULL ;
	mode->mmio      = &MMIO ;

	dpy->mode = *(mode) ;
}

kgi_sint 
kgim_chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			enum kgi_tm_cmd *cmd)
{
  int bpp ; 
  int textx, texty;
  GET_PRIVATE ;
  

  DEBUG1(" ENTER kgim_chipset_check_mode");
 
  /* return -E(CHIPSET,FAILED); */

  textx = mode->request.dpp.x;
  texty = mode->request.dpp.y;

  switch (mode->request.graphtype)
    {
    case KGIGT_8BIT:
      bpp = 8;	
      break;
    case KGIGT_16BIT:
      bpp = 16;
      break;
    case KGIGT_24BIT:
      bpp = 24;
      break;
    case KGIGT_32BIT:
      bpp = 32;
      break ;
    default:
      DEBUG0("graphtype %i not supported", mode->request.graphtype);
      return -E(CHIPSET,FAILED);
    }

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

  switch (*cmd) 
    {
    case CMD_PROPOSE:
      DEBUG1(" => CMD_PROPOSE %dx%d %dx%d %d",XVIRTUAL,YVIRTUAL,XVISIBLE,YVISIBLE,bpp);
      if (mode->request.graphtype == KGIGT_TEXT16) 
	{
	  
	  mode->magnify.x = mode->magnify.y = 1;
	  if (textx != 8) {
	    DEBUG0("Invalid char width (%i)", textx);
	    return -EINVAL;
	  }
	  
	
	  if (XVIRTUAL == 0 && YVIRTUAL == 0)
	    return -EINVAL;
	  
	  if (XVIRTUAL == 0) 
	    XVIRTUAL = TEXT_LEN / (YVIRTUAL * 2);
	  
	  if (YVIRTUAL == 0)
	    YVIRTUAL = TEXT_LEN / (XVIRTUAL * 2);
	  
	  if ((2 * XVIRTUAL * YVIRTUAL) > TEXT_LEN) {
	    DEBUG1("Not enough text memory");
	    return -ENOMEM;
	  }
	  mode->x.width = mode->request.visible.x * textx;
	  mode->y.width = mode->request.visible.y * texty;

	} 
		
		
      else /* mode->request.graphtype != KGIGT_TEXT16 */
	{        

	  roundmode(&mode->request, 8, 1);
	  if ((XVISIBLE>2048) || (YVISIBLE>2048)) 
	    {
	      DEBUG1("The visible screen size is too large: %dx%d",XVISIBLE,YVISIBLE);
	      return -EINVAL;
	    }
	  if ((XVIRTUAL>2048) || (YVIRTUAL>2048)) 
	    {
	      DEBUG1("The virtual screen size is too large: %dx%d",XVIRTUAL,YVIRTUAL);
	      return -EINVAL;
	    }
	  
	  if (YVISIBLE<175)
	    {
	      DEBUG1("The visible height is to small : %d",YVISIBLE);
	      return -EINVAL;
	    }
	  else if (YVISIBLE<350) 
	    mode->magnify.y=2 ;
	  else 
	    mode->magnify.y=1 ;


	  /* Adjust virtual size */
	  if (adjust_virtual(mode)) {
	    DEBUG1("x virtual size %d too big",
		   mode->request.virt.x);
	    return -EINVAL;
	  }

	  if ((((XVIRTUAL * YVIRTUAL * bpp/8) + 
		0xFFFF) & ~0xFFFF) > MEMSIZE) {
	    DEBUG1("Not enough memory to do "
		   "virtual %ix%i at %i bpp",
		   XVIRTUAL, YVIRTUAL, bpp);
	    return -ENOMEM;
	  }	

#if 0
	  mode->x.width = mode->request.visible.x * 
	    mode->magnify.x;
	  mode->y.width = mode->request.visible.y * 
	    mode->magnify.y;
#else
	  mode->x.width = mode->request.visible.x ;
	  mode->y.width = mode->request.visible.y ;
#endif
	}
      return EOK;
      
    case CMD_LOWER:
      if (mode->dclk < CHIPSET_INFO.dclk.min) {
	DEBUG1("DCLK too low (%i)", mode->dclk);
	return -E(CHIPSET,FAILED);
      }
      mode->dclk = (mode->dclk > CHIPSET_INFO.dclk.max)
	? CHIPSET_INFO.dclk.max : mode->dclk;
      
      adjust_timing(mode);
      return EOK;
      
    case CMD_RAISE:
      if (mode->dclk > CHIPSET_INFO.dclk.max) {
	DEBUG1("DCLK too high (%i)", mode->dclk);
	return -E(CHIPSET,FAILED);
      }
      mode->dclk = (mode->dclk < CHIPSET_INFO.dclk.min)
	? CHIPSET_INFO.dclk.min : mode->dclk;
      
      adjust_timing(mode); 

      return EOK;
      
    case CMD_CHECK:
      if (mode->lclk.mul != 1) {
	DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
	return -E(CHIPSET,FAILED);
      }
      switch (mode->lclk.div) {
      case 1:
	break;
      default:
	DEBUG1("Invalid CLK to LCLK divider %d", mode->lclk.div);
	return -E(CHIPSET,FAILED);
      }
      if (mode->request.graphtype == KGIGT_TEXT16) {
	mode->flags |= FL_CHIP_COLORTEXT;
	if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
	    || (textx != 8) || (texty < 1) || (texty > 32) || 
	    ((2 * XVIRTUAL * YVIRTUAL) > TEXT_LEN)
	    || (mode->dclk > CHIPSET_INFO.dclk.max)
	    || (mode->dclk < CHIPSET_INFO.dclk.min))
	  {
	    DEBUG1("HW check failed");
	    return -E(CHIPSET,FAILED);
	  }
	
      } else {
	
	/* There used to be some XVIRTUAL checking here. */
	
	if (((XVIRTUAL * YVIRTUAL * bpp / 8) > MEMSIZE)
	    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
	    || (mode->clk.mul != 1) || (mode->clk.div != 1)
	    || (mode->dclk < CHIPSET_INFO.dclk.min)
	    || (CHIPSET_INFO.dclk.max < mode->dclk)
	    /* LOOKHERE -- allow double pels in > 8bpp modes -- not good code */
	    /*|| (mode->magnify.x == 2 && mode->request.graphtype != KGIGT_8BIT) */ )
	  {
	    NOTICE1("HW check failed\n");
	    return -E(CHIPSET,FAILED);
	  }
      }
      
#define TM(X) ((CHIPSET_INFO.maxdots.x < mode->x.X) || (CHIPSET_INFO.maxdots.y < mode->y.X))
      if (TM(width) || TM(blankstart) || TM(syncstart)
	  || TM(syncend) || TM(blankend) || TM(total)) {
	NOTICE1("Timing check failed");
	DEBUG1("Maxdots are (%d,%d)", CHIPSET_INFO.maxdots.x,
	       CHIPSET_INFO.maxdots.y);
	return -E(CHIPSET,FAILED);
      }
#undef TM
      	
      chipset_setup_mode(dpy, mode);
      adjust_timing(mode);

#if 0 
      {
	struct statetable setup ;
	chipset_compute_mode(dpy,mode,&setup) ; 
	chipset_dump_state(&setup) ; 
	return -E(CHIPSET,FAILED);
      }
#endif 

      return EOK;
      
    default:
      INTERNAL_ERROR;
      return -E(CHIPSET,FAILED);
    }
  
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
  
  return -E(CHIPSET,FAILED);    
}

static int 
setup_memory(struct kgi_display *dpy)
{
  struct chipset_private *private;
  kgiu32 partSize;                /* size of SGRAM chips in Mbits */
  kgiu32  nChips;                   /* # chips of SDRAM/SGRAM */
  kgiu32  dramInit0_strap;    
  kgiu32  dramInit1_strap;    
  kgiu32  dramInit1;
  kgiu32  miscInit1;
  int memType=-1; /* SDRAM or SGRAM */
  int memSize;

  private = KGIM_PRIVATE(dpy, chipset) ; 

  /* determine memory type: SDRAM or SGRAM */
  memType = MEM_TYPE_SGRAM;
  dramInit1_strap = inl(IOBASE+DRAMINIT1);

  if ( dramInit1_strap & SST_MCTL_TYPE_SDRAM ) memType = MEM_TYPE_SDRAM;
  

    /* set memory interface delay values and enable refresh */
  /* these apply to all RAM vendors */
  dramInit1 = 0x0;
  dramInit1 |= 2<<SST_SGRAM_OFLOP_DEL_ADJ_SHIFT;
  dramInit1 |= SST_SGRAM_CLK_NODELAY;
  dramInit1 |= SST_DRAM_REFRESH_EN;
  dramInit1 |= (0x18 << SST_DRAM_REFRESH_VALUE_SHIFT) & SST_DRAM_REFRESH_VALUE;  
  /* determine memory size from strapping pins (dramInit0 and dramInit1) */
  dramInit0_strap = inl(IOBASE+DRAMINIT0);
  
  dramInit0_strap &= SST_SGRAM_TYPE | SST_SGRAM_NUM_CHIPSETS;
  
  if ( memType == MEM_TYPE_SDRAM )
    {
      memSize = 16;
    } 
  else 
    {
      nChips = ((dramInit0_strap & SST_SGRAM_NUM_CHIPSETS) == 0) ? 4 : 8;
      
      if ( (dramInit0_strap & SST_SGRAM_TYPE) == SST_SGRAM_TYPE_8MBIT )  
	{
	  partSize = 8;
	} 
      else if ( (dramInit0_strap & SST_SGRAM_TYPE) == SST_SGRAM_TYPE_16MBIT) 
	{
	  partSize = 16;
	}
      else 
	{
	  ERROR("Invalid sgram type = 0x%lx",
		 (dramInit0_strap & SST_SGRAM_TYPE) << SST_SGRAM_TYPE_SHIFT );
	  return 0 ;
	}
      memSize = (nChips * partSize) / 8;      /* in MBytes */
    }

    NOTICE(" Found %d Mb of %s memory",memSize,(memType==MEM_TYPE_SDRAM)?"SDRAM":"SGRAM");	

    /*
      disable block writes for SDRAM
    */
    miscInit1 = inl(IOBASE+MISCINIT1);
    if ( memType == MEM_TYPE_SDRAM ) 
      {
	miscInit1 |= SST_DISABLE_2D_BLOCK_WRITE;
      }
    miscInit1|=1;
    outl(miscInit1,IOBASE+MISCINIT1);
    
    return memSize KB ; 
}

kgi_sint 
kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{	
  struct chipset_private *private;
  struct pci_dev *pcidev ;
 
  /* 'private' must be initialized first  */
  /* It is used by several macros : IOBASE,CHIPSET_INFO,....) */
  private = kgi_malloc_small(sizeof(*private));
  memset(private, 0, sizeof(*private));

  if (private == NULL) {
    ERROR("ERROR: failed to allocate private data!");
    return -ENOMEM;
  }

  /* now we can work */

  IOBASE=0 ; 
  MEMBASE=0 ; 
  IOMEMBASE=0 ; 

  
  TYPE = none_3dfx ; 
  for (pcidev = pci_devices; pcidev != NULL ; pcidev=pcidev->next)
  {
    if (((pcidev->class>>16) & 0xff) == PCI_BASE_CLASS_DISPLAY)
      {
	if (pcidev->vendor == PCI_VENDOR_ID_3DFX)
	  {
	    switch (pcidev->device) { 
	    case PCI_DEVICE_ID_3DFX_VOODOO:
	    case PCI_DEVICE_ID_3DFX_VOODOO2:	
	      NOTICE("  Unsupported 3dfx PCI device detected.");
	      break;  
	    case PCI_VENDOR_ID_3DFX_BANSHEE: /* 0x0003 */
	      NOTICE("  PCI device detected: 3dfx Banshee.");
	      CHIPSET_INFO = banshee_chipinfo ; /* it's a copy, not a pointer */
	      TYPE = banshee_3dfx ;
	      break ; 
	    case 0x0005: /* 0x0003 */
	      NOTICE("  PCI device detected: 3dfx Voodoo3.");
	      CHIPSET_INFO = banshee_chipinfo ; /* it's a copy, not a pointer */
	      TYPE = banshee_3dfx ;
	      break ; 
	    default:
	      NOTICE(" Unknown 3dfx PCI device detected.");
	    } 
	  } 
	else
	  NOTICE("  PCI device detected: Unknown video card.");
      }
    if (TYPE != none_3dfx)
      break ; 
  }

  if (TYPE==none_3dfx)
    goto error ; 

  KGIM_PRIVATE(dpy, chipset) = private;


  /* install the io region */
  IOBANK = banshee_io_bank  ; 
  pci_read_config_word(pcidev,24,&PIOBASE); PIOBASE&=0xFFFFFFFC;
  IOSIZE = 0x100 ;
#if 1
  IOBASE = PIOBASE  ;
#else
  if (io_check_region(&IOBANK)) 
    goto error ;
#endif
  io_claim_region(&IOBANK);   /* compute IOBASE */
  DEBUG1(" IOBase = %x", IOBASE);
  DEBUG1(" Physical IOBase = %x", PIOBASE);

  RAMSIZE = setup_memory(dpy)*1024 ;
  if (!RAMSIZE)
    goto error ; 

  /* the memory bank 1 contains the video memory. */
  pci_read_config_dword(pcidev,20,&PMEMBASE); PMEMBASE&=0xFFFFFFF0;
  MEMSIZE = RAMSIZE  ;
  if (mem_check_region(&BANK1)) 
    goto error ;
  mem_claim_region(&BANK1);  /* compute MEMBASE */
  DEBUG1(" Memory size = %x", MEMSIZE);
  DEBUG1(" Memory Base = %p", MEMBASE);
  DEBUG1(" Physical Memory Base = %x", PMEMBASE);

  /* the memory bank 0 has a fixed size (for I/O) */
  pci_read_config_dword(pcidev,16,&PIOMEMBASE);PIOBASE&=0xFFFFFFF0;
  IOMEMSIZE = 0x2000000 ;
  if (mem_check_region(&BANK0)) 
    goto error ;
  mem_claim_region(&BANK0);  /* compute MEMBASE */
  DEBUG1(" IO Address = %p", IOMEMBASE);
  DEBUG1(" Physical IO Address = %x", PIOMEMBASE);

  memset(&MMIO, 0, sizeof(MMIO));
  MMIO.next	= NULL;
  MMIO.name	= "Banshee Video RAM";
  MMIO.type	= MMAP_TYPE_MMIO | MMAP_PUBLIC | MMAP_FRAMEBUFFER;
  MMIO.prot	= MMIO_RDWR | MMIO_SHARED;
  MMIO.size     = MEMSIZE ;
  MMIO.psize    = MEMSIZE ;
  MMIO.pstart   = PMEMBASE ;
  MMIO.set_offset = NULL ;
  MMIO.private   = NULL ;

  strcpy(card->manufact,MANUFACT );
  strcpy(card->model, MODEL );
  card->modes = MODES ;
  card->flags = FLAGS ;
  card->maxdots = MAXDOTS; 
  /* card->kgi_monitor = <NOT HERE> */
  /* card->kgi_ramdac = <NOT HERE> */
  /* card->kgi_clock = <NOT HERE> */
  card->chipset = &CHIPSET_INFO;

  /**  card->preset = default_kgi_mode ;  */

  DEBUG1(" EXIT kgim_chipset_init");

  {
    struct statetable state;
    save_state(dpy,&state);
    chipset_dump_state(&state);
  }

  return EOK;

error:
  if (IOBASE)    io_free_region(&IOBANK) ;
  if (IOMEMBASE) mem_free_region(&BANK0) ;
  if (MEMBASE)   mem_free_region(&BANK1) ;
  kgi_free_small(private);
  KGIM_PRIVATE(dpy, chipset) = NULL;
  card->chipset = NULL;
  return -E(CHIPSET,UNKNOWN);
}

#if 0 
void 
banshee_fb_set_split(struct kgi_display *dpy, int line)
{
  register kgiu8 foo;
  IS_CRITICAL;

  GET_PRIVATE;
	
  ENTER_CRITICAL;
  CRT_out8(line & 0xFF, CRT_LINECOMPARE);
  foo = (line & 0x0100)
    ? CRT_in8(0x07) | CR07_LINECOMPARE_B8
    : CRT_in8(0x07) & ~CR07_LINECOMPARE_B8;
  CRT_out8(foo, 0x07);
  
  foo = (line & 0x0200)
    ? CRT_in8(0x09) | CR09_LINECOMPARE_B9
    : CRT_in8(0x09) & ~CR09_LINECOMPARE_B9;
  CRT_out8(foo, 0x09);
  
  foo = (line & 0x0400)
    ? CRT_in8(CRT_OVRFLW_V) | CR35_LINECOMPARE_B10
    : CRT_in8(CRT_OVRFLW_V) & ~CR35_LINECOMPARE_B10;
  CRT_out8(foo, CRT_OVRFLW_V);
  LEAVE_CRITICAL;
}
#endif

int 
kgim_chipset_command(struct kgi_graphic *state, kgi_uint cmd, kgi_uint arg)
{
  kgi_coord *coord;
  struct kgi_font *font;
  struct kgi_display *dpy;

  DEBUG1(" ENTER kgim_chipset_command");

  dpy = state->dev.dpy;
  switch (cmd) {
  case CHIP_SETDEFFONT:
    if (state->dev.mode.graphtype==KGIGT_TEXT16) {
      font = (struct kgi_font *)(state->io_buf);
      text16_set_font(dpy, font, 0, font->info->positions);
      return 0;
    } else return -E(CHIPSET,BOGUS);
  case CHIP_SETSPLITLINE:
    if (arg > state->dev.mode.visible.y)
      return -E(CHIPSET,BOGUS);
    
    if (state->dev.org[0].y > state->dev.mode.visible.y - arg )
      kgi_set_origin(&(state->dev), 0, 
		     state->dev.org[0].x,
		     state->dev.mode.visible.y - arg );
    
    kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
    return 0;
  case CHIP_SETVISFRAME:
    coord = (kgi_coord *)(state->io_buf);
    
    if (coord->x < 0 || 
	coord->x > state->dev.mode.virt.x - state->dev.mode.visible.x)
      return -E(CHIPSET,BOGUS);
    
    if (coord->y < 0 || dpy->mode.magnify.y<1 ||
	coord->y > state->dev.mode.virt.y - state->dev.split/dpy->mode.magnify.y)
      return -E(CHIPSET,BOGUS);
    
    kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
    return 0;
  default:
    break;
  }
  return -E(CHIPSET,UNKNOWN);
}

#define min(A,B) ((A)<(B)?(A):(B))
#define max(A,B) ((A)<(B)?(B):(A))



static void 
chipset_dump_state(struct statetable *state) 
{
  int i ; 
  NOTICE("========== DUMP =========") ;
  NOTICE(" vidcfg     =%x",state->vidcfg) ;
  NOTICE(" vidpll     =%x",state->vidpll);
  NOTICE(" mempll     =%x",state->mempll);
  NOTICE(" gfxpll     =%x",state->gfxpll);
  NOTICE(" dacmode    =%x",state->dacmode);
  NOTICE(" vgainit0   =%x",state->vgainit0);
  NOTICE(" vgainit1   =%x",state->vgainit1);
  NOTICE(" screensize =%x",state->screensize);
  NOTICE(" stride     =%x",state->stride);
  NOTICE(" cursloc    =%x",state->cursloc);
  NOTICE(" startaddr  =%x",state->startaddr);
  NOTICE(" ExtVga[0]  =%x",state->ExtVga[0]);
  NOTICE(" ExtVga[1]  =%x",state->ExtVga[1]);
  for(i = 0; i < NoGRCRegs; i++)
    NOTICE(" GRC[%d]  =%x",i,state->GRC[i]);
  for(i = 0; i < NoSEQRegs; i++)
    NOTICE(" SEQ[%d]  =%x",i,state->SEQ[i]);
  for(i = 0; i < NoATCRegs; i++)
    NOTICE(" ATC[%d]  =%x",i,state->ATC[i]);
  for(i = 0; i < NoCRTRegs; i++)
    NOTICE(" CRT[%d]  =%x",i,state->CRT[i]);  
  NOTICE("=========================") ; 
}

static void
chipset_compute_mode(struct kgi_display *dpy, 
		     struct kgi_mode *mode,
		     struct statetable *setup)
{

  int dbl;
  int bpp,cpp;
  int textx,texty;
  int wd, /* logical width of a framebuffer line (stride ?) */
    hd, /* hor display end */
    hss, /* hor sync start */
    hse, /* hor sync end */
    hbs, /* hor blank start */
    hbe, /* hor blank end */
    ht, /* hor total */
    vd, /* vert display end */
    vss, /* vert sync start */
    vse, /* vert sync end */
    vbs, /* vert blank start */
    vbe, /* vert blank end */
    vt, /* vert total */
    ln; /* line compare */

  GET_PRIVATE;
    
  /* I am not sure for  BANSHEE_MAX_CLOCK/2  */
  /* Is there a bug in the XFree86 server ? */


  /*
   *
   *  
   * INIT VGA registers 
   *
   *
   */
  
 
  DEBUG2("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
	 mode->request.virt.y, mode->request.graphtype);
  
  textx = mode->request.dpp.x;
  texty = mode->request.dpp.y;
  
  /*      Set up bits per pixel and logical screen width
   */
  switch (mode->request.graphtype) {
    
  case KGIGT_TEXT16:
  case KGIGT_8BIT:
  case KGIGT_16BIT:
  case KGIGT_24BIT:
  case KGIGT_32BIT:
    break;
    
  default:
    INTERNAL_ERROR;
    return;
  }
  
  if (mode->request.graphtype == KGIGT_TEXT16) {
    
    bpp     = 16;
    wd = mode->request.virt.x * 2 / 4;
    /* Due to the addressing mode used */

    setup->CRT[0x17] = CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC |
      CR17_CGA_BANKING | CR17_HGC_BANKING;
    
    setup->CRT[0x14] |= CR14_UNDERLINE_MASK;
    
    setup->SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;
      
    setup->GRC[0x05] |= GR5_ODD_EVEN;
    setup->GRC[0x06] |= GR6_CHAIN_OE;

    setup->GRC[0x06] |= GR6_MAP_B8_32;

    setup->SEQ[0x01] |= SR01_8DOT_CHCLK;

    setup->SEQ[0x04] |= 0x02;

    if (textx!=8) {
      INTERNAL_ERROR;
      return;
    }

    setup->ATC[0x13] = 0x00;

  } else { 

    setup->SEQ[0x00] = 0x03;
    setup->SEQ[0x01] = 0x01; /*0x09*/
    setup->SEQ[0x02] = 0x0F;
    setup->SEQ[0x03] = 0x00;
    setup->SEQ[0x04] = 0x0E;

    setup->GRC[0x00] = 0x00;
    setup->GRC[0x01] = 0x00;
    setup->GRC[0x02] = 0x00;
    setup->GRC[0x03] = 0x00;
    setup->GRC[0x04] = 0x00;
    setup->GRC[0x05] = 0x40;
    setup->GRC[0x06] = 0x05;
    setup->GRC[0x07] = 0x0F;
    setup->GRC[0x08] = 0xFF;

    setup->ATC[0x10] = 0x41;
    setup->ATC[0x11] = 0x00;
    setup->ATC[0x12] = 0x0F;
    setup->ATC[0x13] = 0x00;
    setup->ATC[0x14] = 0x00;


    switch(mode->request.graphtype) {

    case KGIGT_8BIT:
      bpp     = 8;
      break;
    case KGIGT_16BIT:
      bpp     = 16;
      break;
    case KGIGT_24BIT:
      bpp     = 24;
      break;
    case KGIGT_32BIT:
      bpp     = 32;
      break;

    default:
      INTERNAL_ERROR;
      return;
    }

    wd    = mode->request.virt.x * bpp / 128;
    texty = mode->magnify.y;
  }

  cpp = bpp/8 ; 

  /*
        **      Set up timing
        */
   
  hd = (mode->x.width >> 3) - 1;
  hbs = mode->x.blankstart >> 3;
  hss = (mode->x.syncstart >> 3);
  hse = (mode->x.syncend >> 3);
  hbe = mode->x.blankend >> 3;
  ht = (mode->x.total >> 3) - 5;

  vd = mode->y.width - 1;
  vbs = mode->y.blankstart;
  vss = mode->y.syncstart;
  vse = mode->y.syncend;
  vbe = mode->y.blankend;
  vt = mode->y.total - 2;

  if (mode->request.graphtype != KGIGT_TEXT16) {
    hbs = (mode->x.width >>3)-1;
    hbe = (mode->x.total >>3)-1;  
    vbs=mode->y.width-1;
    vbe=mode->y.total-1;
  }

  /* Warning: the clock type is set later */
  setup->MISC = ((mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC) 
    | ((mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC) 
    | MISC_COLOR_IO | 0xC;
    ;

   

  setup->CRT[0]  = ht;
  setup->CRT[1]  = hd;
  setup->CRT[2]  = hbs & 0xFF;
  setup->CRT[3]  = (hbe & 0x1F) | CR03_IS_VGA;
  setup->CRT[4]  = hss;
  setup->CRT[5]  = ((hbe & 0x20) << 2) | (hse & 0x1F);
  setup->CRT[6]  = vt & 0xFF;
  setup->CRT[7]  = ((vt & 0x100) >> 8 ) | ((vd & 0x100) >> 7 ) |
    ((vss & 0x100) >> 6 ) | ((vbs & 0x100) >> 5 ) |
    0x10 | ((vt & 0x200) >> 4 ) | 
    ((vd & 0x200) >> 3 ) | ((vss & 0x200) >> 2 );
  setup->CRT[8]  = 0;
  setup->CRT[9]  = 0x40 | ((vbs & 0x200) >> 4) | 
    ((texty - 1) & 0x1F);
  if (mode->magnify.y==2) 
    setup->CRT[9] |= 0x80;
  setup->CRT[16] = vss & 0xFF;
  setup->CRT[17] = (vse & 0x0F) | CR11_DISABLE_VSYNC_IRQ; /* clear too */
  setup->CRT[18] = vd & 0xFF;
  setup->CRT[19] = wd & 0xFF;
  setup->CRT[21] = vbs & 0xFF;
  setup->CRT[22] = vbe & 0xFF;
  setup->CRT[23] = 0xa3;
  setup->CRT[24] = 0xFF;
	
  setup->ExtVga[0]= ((ht&0x100)>>8) | ((hd&0x100)>>6) | 
    ((hbs&0x100)>>4) | ((hss&0x100)>>2);

  setup->ExtVga[1]= ((vt&0x400)>>10) | ((vd&0x400)>>8) | 
    ((vbs&0x400)>>6) | ((vss&0x400)>>4);
      
  /*
   *
   *
   *
   *
   *
   *
   *
   */

  setup->vgainit0 = 0;
  setup->vgainit0 |= SST_VGA0_EXTENSIONS;
  setup->vgainit0 |= SST_WAKEUP_3C3 << SST_VGA0_WAKEUP_SELECT_SHIFT;
  setup->vgainit0 |= SST_VGA0_ENABLE_DECODE << SST_VGA0_LEGACY_DECODE_SHIFT;
  setup->vgainit0 |= SST_ENABLE_ALT_READBACK << SST_VGA0_CONFIG_READBACK_SHIFT;
  setup->vgainit0 |= BIT(12);
  
  setup->vgainit1 = inl(IOBASE+VGAINIT1);
  setup->vgainit1 &= 0x1FFFFF;

  setup->vidcfg = SST_VIDEO_PROCESSOR_EN  
    | SST_DESKTOP_EN 
    | (cpp-1) << SST_DESKTOP_PIXEL_FORMAT_SHIFT;

  if (cpp!=1)
    setup->vidcfg |= SST_DESKTOP_CLUT_BYPASS;

  if (mode->request.graphtype == KGIGT_TEXT16) {
    setup->vgainit0 = 140;
    setup->vgainit1 = 0;
    setup->vidcfg = 0;
  }


  setup->stride = mode->request.virt.x * cpp ;
  setup->startaddr = 0 ; 
/*  setup->dacmode = 0  ;*/
  setup->cursloc = 0 ;

  /*  setup->cursloc = pVB->CursorData; */



  /* should be computed (see X server) !!!!! */
  setup->mempll = 0xf91e /* & 0x7FFF */ ; 
  setup->gfxpll = 0xf91e /* & 0x7FFF */ ;

  /* Computed by the clock driver */
  /* setup->vidpll = VIDPLL ;  */

  /* Set the screen size */
  
  if (mode->magnify.y==2) 
    {
      setup->screensize=mode->x.width|(mode->y.width<<13);
      /*   setup->vidcfg |= SST_HALF_MODE */
    } 
  else 
    {
      setup->screensize=mode->x.width|(mode->y.width<<12);
      setup->vidcfg &= ~SST_HALF_MODE;
    }
  

}




void 
kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{ 
  struct statetable setup = default_state ;

  GET_PRIVATE;
  IS_CRITICAL ;

  DEBUG1("*************** kgim_chipset_set_mode") ;

  if (mode==&dpy->card->preset) 
    {
      DEBUG1("*************** EXIT because of Reset mode") ;
      return ;
    }

#if 0  
  if (mode->dclk > BANSHEE_MAX_CLOCK/2 ) {
    mode->x.width>>=1;
    mode->x.syncstart>>=1;
    mode->x.syncend >>=1;
    mode->x.total>>=1;
    /* mode->CrtcHSkew>>=1; NOT available in GGI ?  */
    dbl=TRUE;
  }
#endif

  chipset_compute_mode(dpy,mode,&setup) ;
 
  ENTER_CRITICAL;
  kgim_monitor_set_mode(dpy, mode);
  kgim_ramdac_set_mode(dpy, mode);
  kgim_clock_set_mode(dpy, mode);
  setup.vidpll = VIDPLL ;  /* computed by kgim_clock_set_mode */
  kgim_accel_set_mode(dpy, mode);
#if 1
  restore_state(dpy, &setup); /* also enables sync signals */
#endif
  LEAVE_CRITICAL;

  chipset_setup_mode(dpy,mode) ;
  /* chipset_dump_state(&setup) ; */
  save_state(dpy, &setup);
  /* chipset_dump_state(&setup) ;  */

#if 0  
  if (dbl) {
    mode->x.total<<=1;
    mode->x.width<<=1;
    mode->x.syncstart<<=1;
    mode->x.syncend<<=1;
    /* mode->CrtcHSkew<<=1; NOT available in GGI ? */ 
  }    
#endif

}

 
