 /*

********************************************************************************
**      ATI Mach 64 chipset driver
********************************************************************************
**
    Copyright (C)       1998    Isaac Connor
                                Mathew Wilkins

********************************************************************************
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************

*/

/*--------
 *	propose video mode
 */
#include <kgi/system.h>
#include <kgi/kgi.h>

#ifndef _MACH64_CORE_C__
#define _MACH64_CORE_C__

#include "ATI/mach64.h"
#include "ATI/mach64_modes.h"

int	mach64_propose_mode( struct kgi_display *dpy, struct kgi_mode *mode )
{
int bpp;
	switch( mode->request.graphtype )
	{
		case	KGIGT_TEXT32:
		case	KGIGT_1BIT:
		case	KGIGT_4BIT:  

	/*   someone else should implement this one, kgicon doesn't do text. */		
		case	KGIGT_TEXT16:

			return -E( CHIPSET, FAILED );

		case	KGIGT_8BIT:   bpp=8; break;
		case	KGIGT_15BIT:  
		case	KGIGT_16BIT:  bpp=16;break;
		case	KGIGT_24BIT:  bpp=24;break;
		case	KGIGT_32BIT:  bpp=32;break;

		default:
			printk( KERN_ERR "mach64_propose_mode: Unknown graphics type.\n" );
			return -E( CHIPSET, FAILED );

	}/* switch */

        #define XVIRTUAL (mode->request.virt.x)
	#define YVIRTUAL (mode->request.virt.y)
	#define XVISIBLE (mode->request.visible.x)
	#define YVISIBLE (mode->request.visible.y)
	
        if (XVISIBLE <= 0 || YVISIBLE <= 0) 
		    return -EINVAL;
	mode->magnify.x = mode->magnify.y = 1;	

	mode->x.width = XVISIBLE;
	mode->y.width = YVISIBLE;
	#undef XVIRTUAL
	#undef YVIRTUAL
	#undef XVISIBLE
	#undef XVISIBLE
	
	return EOK;
	
}

/*--------
 *	check video mode
 */

int	mach64_check_mode( struct kgi_display *dpy, struct kgi_mode *mode )
{
int bpp=0;
	switch( mode->request.graphtype )
	{
		case	KGIGT_TEXT32:
		case	KGIGT_1BIT:
		case	KGIGT_4BIT:

		case	KGIGT_TEXT16:
			return -E( CHIPSET, FAILED );

		case	KGIGT_8BIT:   bpp=8; break; 
		case	KGIGT_15BIT:  
		case	KGIGT_16BIT:  bpp=16;break;
		case	KGIGT_24BIT:  bpp=24;break;
		case	KGIGT_32BIT:  bpp=32;break;

		default:
			printk( KERN_ERR "mach64_check_mode: Unknown graphics type.\n" );
			return -E( CHIPSET, FAILED );

	}/* switch */
		if(mode->request.virt.y == 0)
			mode->request.virt.y = mode->request.visible.y;
		if(mode->request.virt.x == 0)
			mode->request.virt.x = mode->request.visible.x;
		mode->fsize = ( mode->request.virt.x * 
					mode->request.virt.y * bpp /8) ;
		mode->fb_size = (mode->fsize + 0xFFFF) & ~0xFFFF;
		mode->mem=NULL;
		mode->io=NULL;
	return EOK;
}




int	mach64_set_mode_grfx( struct kgi_display *dpy, struct kgi_mode *mode )
{
	return -E( CHIPSET, FAILED );
}




void	mach64_adjust_timing_values_grfx( struct kgi_mode *mode )
{
	/* round values down to a multiple of eight. */
	mode->x.width		&=	~7;
	mode->x.blankstart	&=	~7;
	mode->x.syncstart	&=	~7;
	mode->x.syncend		&=	~7;
	mode->x.blankend	&=	~7;
	mode->x.total		&=	~7;
}




int	mach64_clock_lower_grfx( struct kgi_display *dpy, struct kgi_mode *mode )
{
/* seems like a job for the clock driver. */
/*
	if( mode->dclk < Mach64_Chip_Info.dclk.min )
	{
		printk( KERN_ERR "mach64_clock_lower_grfx: Minimum dotclock limit violated.\n" );
		return -E( CHIPSET, FAILED );
	}

	if( mode->dclk > Mach64_Chip_Info.dclk.max )
	{
		mode->dclk	= Mach64_Chip_Info.dclk.max;
	}

	mach64_adjust_timing_values_grfx( mode );
*/
	return	EOK;

}




int	mach64_clock_lower( struct kgi_display *dpy, struct kgi_mode *mode )
{
	switch( mode->request.graphtype )
	{
		case	KGIGT_TEXT32:
		case	KGIGT_1BIT:
		case	KGIGT_4BIT:

		case	KGIGT_TEXT16:
			return -E( CHIPSET, FAILED );



		case	KGIGT_8BIT:
		case	KGIGT_15BIT:
		case	KGIGT_16BIT:
		case	KGIGT_24BIT:
		case	KGIGT_32BIT:

			return	mach64_clock_lower_grfx( dpy, mode );

		default:
			printk( KERN_ERR "mach64_clock_lower: Unknown graphics type.\n" );
			return -E( CHIPSET, FAILED );
	}
}




int	mach64_clock_raise_grfx( struct kgi_display *dpy, struct kgi_mode * mode )
{
/* leave it to the clock driver. */
/*	
	if( mode->dclk > Mach64_Chip_Info.dclk.max )
	{
		printk( KERN_ERR "mach64_clock_raise_grfx: Maximum dotclock limit violated.\n" );
		return -E( CHIPSET, FAILED );
	}

	if( mode->dclk < Mach64_Chip_Info.dclk.min )
	{
		mode->dclk	= Mach64_Chip_Info.dclk.min;
	}

	mach64_adjust_timing_values_grfx( mode );
*/
	return	EOK;
}




int	mach64_clock_raise( struct kgi_display *dpy, struct kgi_mode *mode )
{
	switch( mode->request.graphtype )
	{
		case	KGIGT_TEXT32:
		case	KGIGT_1BIT:
		case	KGIGT_4BIT:

		case	KGIGT_TEXT16:
			return -E( CHIPSET, FAILED );


		case	KGIGT_8BIT:
		case	KGIGT_15BIT:
		case	KGIGT_16BIT:
		case	KGIGT_24BIT:
		case	KGIGT_32BIT:

			return	mach64_clock_raise_grfx( dpy, mode );

		default:
/*			printk( KERN_ERR "mach64_clock_raise: Unknown graphics type.\n" ); */
			return -E( CHIPSET, FAILED );
	}

}

#endif  /* _MACH64_CORE_C__ */


